import random

import networkx as nx
from deepsnap.hetero_graph import HeteroGraph

import pandas as pd
import numpy as np
import torch

class DISNETConstructor():
    def __init__(self, device='cpu'):
        self.device = device

    def DISNETHeterographDeepSnap(self):
        # DataFrames of each type of nodes
        dis = pd.read_csv('data/nodes/dis.tsv', sep='\t')
        dru = pd.read_csv('data/nodes/dru.tsv', sep='\t')

        # Nodes data pre-processing (mapping and data for the graph)
        nodes_flat = pd.concat([dis, dru],
                               keys=['disorder', 'drug'],
                               names=['node_type', 'NID']).reset_index()

        nodes_flat['node_id'] = nodes_flat.index
        nodes_flat_list = nodes_flat.values.tolist()  # data for the graph

        nodes_mapping = nodes_flat[['NID', 'node_id', 'node_type']].set_index(['NID', 'node_type']).to_dict()[
            'node_id']  # mapping

        # Almacena los tipos y un diccionario de tamaños.
        ntypes = ["disorder", "drug"]
        nsizes = {"disorder": dis.shape[0],
                "drug": dru.shape[0]
        }

        # Adding NID to nodes
        dis['NID'] = dis.index
        dis['node_type'] = 'disorder'
        dis['node_id'] = nodes_flat.loc[nodes_flat['node_type'] == 'disorder'].reset_index(drop=True).node_id

        dru['NID'] = dru.index
        dru['node_type'] = 'drug'
        dru['node_id'] = nodes_flat.loc[nodes_flat['node_type'] == 'drug'].reset_index(drop=True).node_id


        # Nodes dataframes to dict to apply map later
        dis_dict = dis[['id', 'node_id']].set_index('id').to_dict()['node_id']
        dru_dict = dru[['id', 'node_id']].set_index('id').to_dict()['node_id']


        # Add nodes to the graph
        G = nx.DiGraph()

        # Node embedding
        dru_embed = torch.load('data/features/dru.pt', map_location=self.device)
        i = 0

        for node in nodes_flat_list:
            if node[0] == 'drug':
                node_feature = dru_embed[i]
                i += 1
            else:
                node_feature = torch.tensor([1] * 100, dtype=torch.float32)
            #node_feature = torch.tensor([1] * 100, dtype=torch.float32)
            G.add_node(node[4], node_type=node[0], node_feature=node_feature, node_label=node[1])

        # -----------------------------
        # LINKS DATA
        # -----------------------------

        # DataFrames of each type of edges
        dis_dru_the = pd.read_csv('data/links/dis_dru_the.tsv', sep='\t')
        dse_sym = pd.read_csv('data/links/dse_sym.tsv', sep='\t')

        # Convert ids to NIDs
        dis_dru_the['disNID'] = dis_dru_the.dis.map(dis_dict)
        dis_dru_the['druNID'] = dis_dru_the.dru.map(dru_dict)

        dse_sym['dseNID'] = dse_sym.dse.map(dis_dict)
        dse_sym['symNID'] = dse_sym.sym.map(dis_dict)

        edges = {
                'dis_dru_the': dis_dru_the[['disNID', 'druNID']].values.tolist(),
                'dru_dis_the': dis_dru_the[['druNID', 'disNID']].values.tolist(),

                'dse_sym': dse_sym[['dseNID', 'symNID']].values.tolist(),
                'sym_dse': dse_sym[['symNID', 'dseNID']].values.tolist()
                }

        edges_dict = {'dis_dru_the': ('disorder', 'drug'),
                      'dru_dis_the': ('drug', 'disorder'),
                      'dse_sym': ('disorder', 'disorder'),
                      'sym_dse': ('disorder', 'disorder')
                      }

        for edge_t in edges_dict.keys():
            for edge in edges[edge_t]:
                try:
                    G.add_edge(int(edge[0]), int(edge[1]), edge_feature=edge[2], edge_type=edge_t)
                except:
                    G.add_edge(int(edge[0]), int(edge[1]), edge_type=edge_t)

        # --------------------------
        # HETEROGRAPGH
        # ------------------------
        hetero = HeteroGraph(G)


        print("The DISNET DEEPSNAP HETEROGRAPH has been generated! :)")
        return hetero

    def decodePredictions(self, list, type, n, prepared=False, name=''):
        dir = 'results/'+type+'_'+str(n)+ name + '_'+'table.csv'
        if(type == 'dis_dru_the'):
            #Read the data
            h = pd.read_csv('data/nodes/dis.tsv', sep='\t')
            t = pd.read_csv('data/nodes/dru.tsv', sep='\t')

            #Give each node an ID
            h['NID'] = h.index
            t['NID'] = t.index
            # Nodes dataframes to dict to apply map later
            head_dict = h[['id', 'NID']].set_index('id').to_dict()['NID']
            tail_dict = t[['id', 'NID']].set_index('id').to_dict()['NID']

        decoded = []
        vals = []
        if not prepared:
            for i,(head,tails,preds) in enumerate(list):
                decoded.append([[head,item.item()] for item in tails])
                vals.append(preds)

                if(i%1000==0 ):
                    print("         Decoded",i,"elements.")

            #Flatten both lists
            decoded = [item for sublist in decoded for item in sublist]
            vals = [item for sublist in vals for item in sublist]

        else:
            decoded = list
            vals = [item[2] for item in list]
   
        vals = np.array(vals)
        index = np.argsort(vals)
        
        index = np.flip(index)
        topN = index[:n]

        complete = []
        for i in topN:
            headData = h.iloc[int(decoded[i][0])]
            tailData = t.iloc[int(decoded[i][1])]
    
            complete.append([headData[0],headData[1],tailData[0],tailData[1],vals[i]])
             
        complete = np.array(complete)
        headType, tailType, _ = type.split("_", 2)
        df = pd.DataFrame(
            {headType: complete[:,0],
             headType + ' name': complete[:,1],
             tailType: complete[:,2],
             tailType + ' name': complete[:,3],
             'pred': complete[:,4]
             })
        print("Table for", type, "new edge predictions:")
        pd.set_option("display.max_rows", n, "display.max_columns", None, 'expand_frame_repr', False)
        df2 = pd.concat([df[:50],df[-50:]])
        print(df2)
        print("    Table saved as:", dir)
        df.to_csv('./'+dir)
        return df

    def graphWithoutRepoDBInferenceSnap(self):
        # DataFrames of each type of nodes
        dis = pd.read_csv('data/nodes/dis.tsv', sep='\t')
        dru = pd.read_csv('data/nodes/dru.tsv', sep='\t')

        # Nodes data pre-processing (mapping and data for the graph)
        nodes_flat = pd.concat([dis, dru],
                               keys=['disorder', 'drug'],
                               names=['node_type', 'NID']).reset_index()

        nodes_flat['node_id'] = nodes_flat.index
        nodes_flat_list = nodes_flat.values.tolist()  # data for the graph

        nodes_mapping = nodes_flat[['NID', 'node_id', 'node_type']].set_index(['NID', 'node_type']).to_dict()[
            'node_id']  # mapping

        # Almacena los tipos y un diccionario de tamaños.
        ntypes = ["disorder", "drug"]
        nsizes = {"disorder": dis.shape[0],
                  "drug": dru.shape[0]
                  }

        # Adding NID to nodes
        dis['NID'] = dis.index
        dis['node_type'] = 'disorder'
        dis['node_id'] = nodes_flat.loc[nodes_flat['node_type'] == 'disorder'].reset_index(drop=True).node_id

        dru['NID'] = dru.index
        dru['node_type'] = 'drug'
        dru['node_id'] = nodes_flat.loc[nodes_flat['node_type'] == 'drug'].reset_index(drop=True).node_id

        # Nodes dataframes to dict to apply map later
        dis_dict = dis[['id', 'node_id']].set_index('id').to_dict()['node_id']
        dru_dict = dru[['id', 'node_id']].set_index('id').to_dict()['node_id']

        # Add nodes to the graph
        G = nx.DiGraph()

        # Node embedding
        dru_embed = torch.load('data/features/dru.pt', map_location=self.device)
        i = 0

        for node in nodes_flat_list:
            if node[0] == 'drug':
                node_feature = dru_embed[i]
                i += 1
            else:
                node_feature = torch.tensor([1] * 100, dtype=torch.float32)
            #node_feature = torch.tensor([1] * 100, dtype=torch.float32)
            G.add_node(node[4], node_type=node[0], node_feature=node_feature, node_label=node[1])

        # -----------------------------
        # LINKS DATA
        # -----------------------------

        # DataFrames of each type of edges
        dis_dru_the = pd.read_csv('data/links/dis_dru_the.tsv', sep='\t')
        dis_dru_the_repoDBAll = pd.read_csv('testData/drugdis_repodb_ALLlinks.tsv', sep='\t')
        dse_sym = pd.read_csv('data/links/dse_sym.tsv', sep='\t')

        # Convert ids to NIDs
        dis_dru_the['disNID'] = dis_dru_the.dis.map(dis_dict)
        dis_dru_the['druNID'] = dis_dru_the.dru.map(dru_dict)

        dis_dru_the_repoDBAll['disNID'] = dis_dru_the_repoDBAll.dis.map(dis_dict)
        dis_dru_the_repoDBAll['druNID'] = dis_dru_the_repoDBAll.dru.map(dru_dict)


        dse_sym['dseNID'] = dse_sym.dse.map(dis_dict)
        dse_sym['symNID'] = dse_sym.sym.map(dis_dict)


        #Delete from the graph the edges of repo_db
        dis_dru_the = pd.concat([dis_dru_the, dis_dru_the_repoDBAll])
        dis_dru_the.drop_duplicates(keep=False, inplace=True)

        dis_dru_the = dis_dru_the[:50355]


        dis_dru_the_repoDBAll = (
            torch.tensor(dis_dru_the_repoDBAll['disNID'].astype(np.int32).to_numpy(), dtype=torch.int32, device=self.device),
            torch.tensor(dis_dru_the_repoDBAll['druNID'].astype(np.int32).to_numpy()-len(dis_dict), dtype=torch.int32, device=self.device))


        edges = {
                'dis_dru_the': dis_dru_the[['disNID', 'druNID']].values.tolist(),
                'dru_dis_the': dis_dru_the[['druNID', 'disNID']].values.tolist(),

                'dse_sym': dse_sym[['dseNID', 'symNID']].values.tolist(),
                'sym_dse': dse_sym[['symNID', 'dseNID']].values.tolist()
                }

        edges_dict = {'dis_dru_the': ('disorder', 'drug'),
                      'dru_dis_the': ('drug', 'disorder'),
                      'dse_sym': ('disorder', 'disorder'),
                      'sym_dse': ('disorder', 'disorder')
                      }

        for edge_t in edges_dict.keys():
            for edge in edges[edge_t]:
                try:
                    G.add_edge(int(edge[0]), int(edge[1]), edge_feature=edge[2], edge_type=edge_t)
                except:
                    G.add_edge(int(edge[0]), int(edge[1]), edge_type=edge_t)

        # --------------------------
        # HETEROGRAPGH
        # ------------------------
        hetero = HeteroGraph(G)

        print("The DISNET DEEPSNAP HETEROGRAPH (without RepoDB edges) has been generated! :)")
        return hetero, dis_dru_the_repoDBAll

    def allRepoDB(self):
        # DataFrames of each type of nodes
        dis = pd.read_csv('data/nodes/dis.tsv', sep='\t')
        dru = pd.read_csv('data/nodes/dru.tsv', sep='\t')

        # Nodes data pre-processing (mapping and data for the graph)
        nodes_flat = pd.concat([dis, dru],
                               keys=['disorder', 'drug'],
                               names=['node_type', 'NID']).reset_index()

        nodes_flat['node_id'] = nodes_flat.index
        nodes_flat_list = nodes_flat.values.tolist()  # data for the graph

        nodes_mapping = nodes_flat[['NID', 'node_id', 'node_type']].set_index(['NID', 'node_type']).to_dict()[
            'node_id']  # mapping


        # Adding NID to nodes
        dis['NID'] = dis.index
        dis['node_type'] = 'disorder'
        dis['node_id'] = nodes_flat.loc[nodes_flat['node_type'] == 'disorder'].reset_index(drop=True).node_id

        dru['NID'] = dru.index
        dru['node_type'] = 'drug'
        dru['node_id'] = nodes_flat.loc[nodes_flat['node_type'] == 'drug'].reset_index(drop=True).node_id

        # Nodes dataframes to dict to apply map later
        dis_dict = dis[['id', 'node_id']].set_index('id').to_dict()['node_id']
        dru_dict = dru[['id', 'node_id']].set_index('id').to_dict()['node_id']


        # -----------------------------
        # LINKS DATA
        # -----------------------------

        # DataFrames of each type of edges
        dis_dru_the_repoDB = pd.read_csv('testData/drugdis_repodb_ALLlinks.tsv', sep='\t')

        # Convert ids to NIDs
        dis_dru_the_repoDB['disNID'] = dis_dru_the_repoDB.dis.map(dis_dict)
        dis_dru_the_repoDB['druNID'] = dis_dru_the_repoDB.dru.map(dru_dict) 

        dis_dru_the_repoDB = (
            torch.tensor(dis_dru_the_repoDB['disNID'].astype(np.int32).to_numpy(), dtype=torch.int32, device=self.device),
            torch.tensor(dis_dru_the_repoDB['druNID'].astype(np.int32).to_numpy()-len(dis_dict), dtype=torch.int32, device=self.device))

        return dis_dru_the_repoDB
        
    def getNid(self,elem,type):
        if type == 'drug':
            dru = pd.read_csv('data/nodes/dru.tsv', sep='\t')
            dru['NID'] = dru.index
            dru_dict = dru[['id', 'NID']].set_index('id').to_dict()['NID']
            index = dru_dict[elem]

            data = dru.iloc[index]
            return index, data[1]

        elif type == 'disease':
            dis = pd.read_csv('data/nodes/dis.tsv', sep='\t')
            dis['NID'] = dis.index
            dis_dict = dis[['id', 'NID']].set_index('id').to_dict()['NID']
            index = dis_dict[elem]

            data = dis.iloc[index]
            return index, data[1]